@@ -3,7 +3,7 @@
 # vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
 # Author: "Chris Ward" <cward@redhat.com>
 
-from __future__ import unicode_literals
+from __future__ import unicode_literals, absolute_import
 
 from ._version import __version__, version_info
 # touch it to avoid pep8 error 'imported but unused'
@@ -11,12 +11,13 @@ __version__, version_info
 
 # setup default root logger
 import logging
-log_format = "%(name)s.%(process)s:%(asctime)s:%(message)s"
+log_format = "%(levelname)s.%(name)s.%(process)s:%(asctime)s:%(message)s"
 logging.basicConfig(format=log_format)
 logger = logging.getLogger()
 logger.setLevel(logging.WARN)
 logger = logging.getLogger('metrique')
 logger.setLevel(logging.WARN)
+
 import os
 # if HOME environment variable is set, use that
 # useful when running 'as user' with root (supervisord)
@@ -32,6 +33,20 @@ os.environ['METRIQUE_TMP'] = os.environ.get(
     'METRIQUE_TMP', os.path.join(prefix, 'tmp'))
 os.environ['METRIQUE_CACHE'] = os.environ.get(
     'METRIQUE_CACHE', os.path.join(prefix, 'cache'))
+os.environ['METRIQUE_STATIC'] = os.environ.get(
+    'METRIQUE_STATIC', os.path.join(prefix, 'static'))
+# if this isn't set in env already, we can only guess...
+# first guess is top-level directory holding the metrique module
+_mod_file = os.path.abspath(__file__)
+_cwd = os.path.abspath('.')
+_src = _cwd.replace(_mod_file, '')
+_src_git = os.path.join(_src, '.git')
+if not (os.path.exists(_src) and os.path.exists(_src_git)):
+    _src = os.path.abspath('.')
+    # last guess is just 'current working directory'
+os.environ['METRIQUE_SRC'] = os.environ.get(
+    'METRIQUE_SRC', _src)
+
 
 # Don't locale.setlocale(); ... "Setting system default encoding is a
 # bad idea because some modules and libraries you use can
@@ -45,12 +60,7 @@ os.environ['METRIQUE_CACHE'] = os.environ.get(
 # sys.stdout = codecs.getwriter('utf8')(sys.stdout)
 
 # ATTENTION: this is the main interface for clients!
-from metrique.core_api import BaseClient as pyclient
+from metrique.metrique import Metrique as pyclient
 from metrique.core_api import MetriqueContainer, MetriqueObject
-from metrique.core_api import MongoDBContainer, MongoDBProxy
-from metrique.core_api import SQLAlchemyContainer, SQLAlchemyProxy
-from metrique.core_api import SQLAlchemyMQLParser
 # avoid lint 'defined by not used' error
-pyclient, MetriqueObject, MetriqueContainer
-MongoDBProxy, MongoDBContainer,
-SQLAlchemyProxy, SQLAlchemyContainer, SQLAlchemyMQLParser
+pyclient, MetriqueContainer, MetriqueObject
