@@ -5,31 +5,51 @@
 
 '''
 metrique.result
-~~~~~~~~~~~~~~~~~
+~~~~~~~~~~~~~~~
 
 This module contains a Pandas DataFrame wrapper and
 additional Pandas object helper functions which is
 used to load and manipulate cube objects.
 '''
 
-from __future__ import unicode_literals
+from __future__ import unicode_literals, absolute_import
 
-from decorator import decorator
-from datetime import timedelta
 import logging
-import numpy as np
-from pandas import DataFrame, Series
-import pandas.tseries.offsets as off
-from pandas.tslib import Timestamp
-import pandas as pd
+logger = logging.getLogger('metrique')
+
+try:
+    from decorator import decorator
+    HAS_DECORATOR = True
+except ImportError:
+    HAS_DECORATOR = False
+    logger.warn('decorator module is not installed')
+
+
+from datetime import timedelta
+
+try:
+    import numpy as np
+    NUMPY_NUMERICAL = [np.float16, np.float32, np.float64, np.float128,
+                       np.int8, np.int16, np.int32, np.int64]
+    HAS_NUMPY = True
+except ImportError:
+    logger.warn('numpy module is not installed')
+    NUMPY_NUMERICAL = []
+    HAS_NUMPY = False
+
+try:
+    from pandas import DataFrame, Series
+    import pandas.tseries.offsets as off
+    from pandas.tslib import Timestamp
+    import pandas as pd
+    HAS_PANDAS = True
+except ImportError:
+    HAS_PANDAS = False
+    DataFrame = object  # load as object, we'll fail at runtime
+    logger.warn('pandas module is not installed')
 
 from metrique.utils import dt2ts, utcnow, ts2dt
 
-logger = logging.getLogger(__name__)
-
-NUMPY_NUMERICAL = [np.float16, np.float32, np.float64, np.float128,
-                   np.int8, np.int16, np.int32, np.int64]
-
 
 def filtered(f):
     '''
@@ -37,6 +57,7 @@ def filtered(f):
     dataframes, such that the dataframe is filtered
     according to left and right bounds set.
     '''
+
     def _filter(f, self, *args, **kwargs):
         frame = f(self, *args, **kwargs)
         ret = type(self)(frame)
@@ -44,13 +65,29 @@ def filtered(f):
         ret._rbound = self._rbound
         return ret
 
-    return decorator(_filter, f)
+    if HAS_DECORATOR:
+        return decorator(_filter, f)
+    else:
+        def err_func(*args, **kwargs):
+            raise RuntimeError("`pip install decorator` required")
+        return err_func
 
 
 class Result(DataFrame):
     ''' Custom DataFrame implementation for Metrique '''
     def __init__(self, data=None, date=None):
+        if not HAS_PANDAS:
+            raise RuntimeError("`pip install pandas` required")
+        if not HAS_NUMPY:
+            raise RuntimeError("`pip install numpy` required")
         super(Result, self).__init__(data)
+        if data is not None and len(data) and \
+                any([c not in self for c in ['_oid', '_start', '_end']]):
+            raise ValueError('_oid, _start and _end must be defined')
+        if data is None or not len(data):
+            # add the essential columns
+            for c in ['_oid', '_start', '_end']:
+                self[c] = []
         # The converts are here so that None is converted to NaT
         self.to_datetime('_start')
         self.to_datetime('_end')
@@ -61,7 +98,8 @@ class Result(DataFrame):
             self._lbound = self._rbound = None
             self.set_date_bounds(date)
         # optimization:
-        self._end_isnull = self._end.isnull()
+        if '_end' in self.columns:
+            self._end_isnull = self._end.isnull()
 
     def to_datetime(self, column):
         '''
@@ -267,7 +305,7 @@ class Result(DataFrame):
 
     def persistent_oid_counts(self, dates):
         '''
-        Counts have many objects (identified by their oids) existed before
+        Counts how many objects (identified by their oids) existed before
         or on a given date.
 
         :param dates: list of the dates the count should be computed.
@@ -302,22 +340,37 @@ class Result(DataFrame):
 
         :param index: name of the new column.
         '''
-        def prep(df):
-            ends = set(df._end.tolist())
-            end = pd.NaT if pd.NaT in ends else max(ends)
-            if end is pd.NaT:
-                age = cut_ts - df._start.min()
-            else:
-                age = min(cut_ts, end) - df._start.min()
-            # for some reason this is not working:
-            #last = df[df._end == end].copy()
-            # but this is:
-            last = df[df._end.isin([end])].copy()
-            last[col_name] = age - timedelta(microseconds=age.microseconds)
-            return last
+        min_start_map = {}
+        max_start_map = {}
+        max_start_ser_map = {}
 
-        cut_ts = self._rbound or utcnow()
-        res = pd.concat([prep(df) for _, df in self.groupby(self._oid)])
+        cols = self.columns.tolist()
+        i_oid = cols.index('_oid')
+        i_start = cols.index('_start')
+        i_end = cols.index('_end')
+        for row in self.values:
+            s = row[i_start]
+            oid = row[i_oid]
+            mins = min_start_map.get(oid, s)
+            if s <= mins:
+                min_start_map[oid] = s
+            maxs = max_start_map.get(oid, s)
+            if s >= maxs:
+                max_start_map[oid] = s
+                max_start_ser_map[oid] = row
+
+        vals = max_start_ser_map.values()
+        cut_ts = utcnow()
+        ages = []
+        for row in vals:
+            end = row[i_end]
+            end = cut_ts if end is pd.NaT else min(cut_ts, end)
+            age = end - min_start_map[row[i_oid]]
+            age = age - timedelta(microseconds=age.microseconds)
+            ages.append(age)
+
+        res = pd.DataFrame(max_start_ser_map.values(), columns=cols)
+        res[col_name] = pd.Series(ages, index=res.index)
         return res
 
     @filtered
@@ -328,17 +381,32 @@ class Result(DataFrame):
         Chain is a series of consecutive versions where
         `_end` of one is `_start` of another.
         '''
-        def prep(df):
-            ends = df._end.tolist()
-            maxend = pd.NaT if pd.NaT in ends else max(ends)
-            ends = set(df._end.tolist()) - set(df._start.tolist() + [maxend])
-            if len(ends) == 0:
-                return df
-            else:
-                cutoff = max(ends)
-                return df[df._start > cutoff]
+        cols = self.columns.tolist()
+        i_oid = cols.index('_oid')
+        i_start = cols.index('_start')
+        i_end = cols.index('_end')
 
-        return pd.concat([prep(df) for _, df in self.groupby(self._oid)])
+        start_map = {}
+        end_map = {}
+        for row in self.values:
+            oid = row[i_oid]
+            if oid not in start_map:
+                start_map[oid] = set()
+                end_map[oid] = set()
+            start_map[oid].add(row[i_start])
+            end_map[oid].add(row[i_end])
+
+        cutoffs = {}
+        for oid in start_map:
+            maxend = pd.NaT if pd.NaT in end_map[oid] else max(end_map[oid])
+            ends = end_map[oid] - start_map[oid] - set([maxend])
+            cutoffs[oid] = None if len(ends) == 0 else max(ends)
+
+        vals = [row for row in self.values
+                if cutoffs[row[i_oid]] is None
+                or cutoffs[row[i_oid]] < row[i_start]]
+
+        return pd.DataFrame(vals, columns=cols)
 
     @filtered
     def one_version(self, index=0):
